/*
Copyright (c) 2003-2006, Vrije Universiteit
All rights reserved.

Redistribution and use of the GLOBULE system in source and binary forms,
with or without modification, are permitted provided that the following
conditions are met:

   * Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.

   * Redistributions in binary form must reproduce the above
     copyright notice, this list of conditions and the following
     disclaimer in the documentation and/or other materials provided
     with the distribution.

   * Neither the name of Vrije Universiteit nor the names of the software
     authors or contributors may be used to endorse or promote
     products derived from this software without specific prior
     written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS, AUTHORS, AND
CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL VRIJE UNIVERSITEIT OR ANY AUTHORS OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

This product includes software developed by the Apache Software Foundation
<http://www.apache.org/>.
*/
/*==========================================================================
 * ap_fake.c
 * ---------
 * We currently link the auditor against libglobule to use its util classes,
 * in particular AprError. However, libglobule references to several Apache
 * functions which won't be found when we link the auditor (standalone).
 * Hence this file: it should contain all Apache functions used in Globule
 * and provide a dummy implementation for them.
 *
 * One can get rid of this file by creating a separate library with all
 * Globule's util classes and link the auditor against just that.
 * I deem this not forth the effort at this time, in particular, as I expect
 * some problems with the util classes that may require copy & paste for
 * each different module that uses them (see globule/alloc/GlobuleStlAlloc).
 *
 * Arno, 2003-09-29
 *===========================================================================
 */

#include <httpd.h>
#include <http_config.h>
#include <http_protocol.h>
#include <http_core.h>
#include <http_log.h>

/**
 * Send an entire file to the client, using sendfile if supported by the 
 * current platform
 * @param fd The file to send.
 * @param r The current request
 * @param offset Offset into the file to start sending.
 * @param length Amount of data to send
 * @param nbytes Amount of data actually sent
 * @deffunc apr_status_t ap_send_fd(apr_file_t *fd, request_rec *r, apr_off_t offset, apr_size_t length, apr_size_t *nbytes);
 */
apr_status_t ap_send_fd(apr_file_t *fd, request_rec *r, apr_off_t offset, 
                               apr_size_t length, apr_size_t *nbytes)
{
}


/* ap_hook_handler def
 * Generated by just precompiling a file that includes http_config.h */
void ap_hook_handler(ap_HOOK_handler_t *pf, const char * const *aszPre, const char * const *aszSucc, int nOrder)
{
}


/**
 * Set the content type for this request (r->content_type). 
 * @param r The current request
 * @param ct The new content type
 * @deffunc void ap_set_content_type(request_rec *r, const char* ct)
 * @warning This function must be called to set r->content_type in order 
 * for the AddOutputFilterByType directive to work correctly.
 */
void ap_set_content_type(request_rec *r, const char *ct)
{
}



/**
 * Set the last modified time for the file being sent
 * @param r The current request
 * @deffunc void ap_set_last_modified(request_rec *r)
 */
void ap_set_last_modified(request_rec *r)
{
}



/**
 * The second of the primary logging routines in Apache.  This uses 
 * a printf-like format to log messages to the error_log.
 * @param file The file in which this function is called
 * @param line The line number on which this function is called
 * @param level The level of this error message
 * @param status The status code from the previous command
 * @param p The pool which we are logging for
 * @param fmt The format string
 * @param ... The arguments to use to fill out fmt.
 * @tip Use APLOG_MARK to fill out file and line
 * @warning It is VERY IMPORTANT that you not include any raw data from 
 * the network, such as the request-URI or request header fields, within 
 * the format string.  Doing so makes the server vulnerable to a 
 * denial-of-service attack and other messy behavior.  Instead, use a 
 * simple format string like "%s", followed by the string containing the 
 * untrusted data.
 * @deffunc void ap_log_perror(const char *file, int line, int level, apr_status_t status, apr_pool_t *p, const char *fmt, ...) 
 */
void ap_log_perror(const char *file, int line, int level, 
                             apr_status_t status, apr_pool_t *p, 
                             const char *fmt, ...)
{
}



/**
 * Get the current server name from the request
 * @param r The current request
 * @return the server name
 * @deffunc const char *ap_get_server_name(request_rec *r)
 */
const char *ap_get_server_name(request_rec *r)
{
}



/**
 * Write a buffer for the current request
 * @param buf The buffer to write
 * @param nbyte The number of bytes to send from the buffer
 * @param r The current request
 * @return The number of bytes sent
 * @deffunc int ap_rwrite(const void *buf, int nbyte, request_rec *r)
 */
int ap_rwrite(const void *buf, int nbyte, request_rec *r)
{
}


/**
 * Get the current server port
 * @param The current request
 * @return The server's port
 * @deffunc apr_port_t ap_get_server_port(const request_rec *r)
 */
apr_port_t ap_get_server_port(const request_rec *r)
{
}



/* Used for constructing self-referencing URLs, and things like SERVER_PORT,
 * and SERVER_NAME.
 */
/**
 * build a fully qualified URL from the uri and information in the request rec
 * @param p The pool to allocate the URL from
 * @param uri The path to the requested file
 * @param r The current request
 * @return A fully qualified URL
 * @deffunc char *ap_construct_url(apr_pool_t *p, const char *uri, request_rec *r)
 */
char * ap_construct_url(apr_pool_t *p, const char *uri, request_rec *r)
{
}



/**
 * Retrieve the document root for this server
 * @param r The current request
 * @warning Don't use this!  If your request went through a Userdir, or 
 * something like that, it'll screw you.  But it's back-compatible...
 * @return The document root
 * @deffunc const char *ap_document_root(request_rec *r)
 */
AP_DECLARE(const char *) ap_document_root(request_rec *r)
{
}



/**
 * One of the primary logging routines in Apache.  This uses a printf-like
 * format to log messages to the error_log.
 * @param file The file in which this function is called
 * @param line The line number on which this function is called
 * @param level The level of this error message
 * @param status The status code from the previous command
 * @param s The server on which we are logging
 * @param fmt The format string
 * @param ... The arguments to use to fill out fmt.
 * @tip Use APLOG_MARK to fill out file and line
 * @warning It is VERY IMPORTANT that you not include any raw data from 
 * the network, such as the request-URI or request header fields, within 
 * the format string.  Doing so makes the server vulnerable to a 
 * denial-of-service attack and other messy behavior.  Instead, use a 
 * simple format string like "%s", followed by the string containing the 
 * untrusted data.
 * @deffunc void ap_log_error(const char *file, int line, int level, apr_status_t status, const server_rec *s, const char *fmt, ...) 
 */
AP_DECLARE(void) ap_log_error(const char *file, int line, int level, 
                             apr_status_t status, const server_rec *s, 
                             const char *fmt, ...)
                            __attribute__((format(printf,6,7)))
{
}


/* During the course of debugging I expanded this macro out, so
 * rather than remove all the useful information there is in the
 * following lines, I'm going to leave it here in case anyone
 * else finds it useful.
 *
 * Ben has looked at it and thinks it correct :)
 */
AP_DECLARE(int) ap_hook_post_config(ap_HOOK_post_config_t *pf,
                                    const char * const *aszPre,
                                    const char * const *aszSucc,
                                    int nOrder)
{
}


unixd_config_rec unixd_config;



/arno/apache/master/modules/libglobule.so: undefined reference to
`unixd_config'
/arno/apache/master/modules/libglobule.so: undefined reference to
`ap_get_brigade'
/arno/apache/master/modules/libglobule.so: undefined reference to
`ap_file_walk'
/arno/apache/master/modules/libglobule.so: undefined reference to
`ap_check_cmd_context'
/arno/apache/master/modules/libglobule.so: undefined reference to
`ap_directory_walk'
/arno/apache/master/modules/libglobule.so: undefined reference to
`ap_hook_child_init'
/arno/apache/master/modules/libglobule.so: undefined reference to
`unixd_set_global_mutex_perms'
/arno/apache/master/modules/libglobule.so: undefined reference to
`ap_setup_client_block'
/arno/apache/master/modules/libglobule.so: undefined reference to
`ap_log_assert'
/arno/apache/master/modules/libglobule.so: undefined reference to
`unixd_setup_child'
/arno/apache/master/modules/libglobule.so: undefined reference to
`ap_server_root'
/arno/apache/master/modules/libglobule.so: undefined reference to
`ap_hook_log_transaction'
/arno/apache/master/modules/libglobule.so: undefined reference to
`ap_pass_brigade'
/arno/apache/master/modules/libglobule.so: undefined reference to
`ap_listeners'
/arno/apache/master/modules/libglobule.so: undefined reference to
`ap_get_local_host'
/arno/apache/master/modules/libglobule.so: undefined reference to
`ap_set_content_length'
/arno/apache/master/modules/libglobule.so: undefined reference to
`ap_getword_white_nc'
/arno/apache/master/modules/libglobule.so: undefined reference to
`ap_server_argv0'
/arno/apache/master/modules/libglobule.so: undefined reference to
`ap_log_error'
/arno/apache/master/modules/libglobule.so: undefined reference to
`ap_allow_standard_methods'
/arno/apache/master/modules/libglobule.so: undefined reference to
`ap_set_listener'
/arno/apache/master/modules/libglobule.so: undefined reference to
`ap_get_client_block'
/arno/apache/master/modules/libglobule.so: undefined reference to
`ap_create_per_dir_config'
/arno/apache/master/modules/libglobule.so: undefined reference to
`ap_walk_config'
/arno/apache/master/modules/libglobule.so: undefined reference to `ap_rprintf'
/arno/apache/master/modules/libglobule.so: undefined reference to
`ap_allow_methods'
/arno/apache/master/modules/libglobule.so: undefined reference to
`ap_discard_request_body'
/arno/apache/master/modules/libglobule.so: undefined reference to
`ap_hook_translate_name'
collect2: ld returned 1 exit status
gmake[1]: *** [auditor] Error 1
gmake[1]: Leaving directory
`/home1/arno/dev/apache/pSodium/src/psodium/psodium'
gmake: *** [all-recursive] Error 1
